@@ -12,12 +12,12 @@ from bhfutils.crawler.playwright.cursor.do_async import GhostCursor, create_curs
 from typing import Union, Awaitable, Literal, Callable, Dict, Generator, Optional, Tuple, Type, TypeVar
 
 from playwright.async_api import (
-    BrowserContext,
     Page,
     PlaywrightContextManager,
     Request as PlaywrightRequest,
     Response as PlaywrightResponse,
-    Route,
+    TimeoutError as PlaywrightTimeoutError,
+    Route
 )
 from scrapy import Spider, signals
 from scrapy.crawler import Crawler
@@ -92,19 +92,12 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
 
         # browser
         self.browser_type: str = settings.get("PLAYWRIGHT_BROWSER_TYPE") or "chromium"
-        self.browser_launch_lock = asyncio.Lock()
         self.launch_options: dict = settings.getdict("PLAYWRIGHT_LAUNCH_OPTIONS") or {}
 
         # contexts
-        self.max_pages_per_context: int = settings.getint(
-            "PLAYWRIGHT_MAX_PAGES_PER_CONTEXT"
-        ) or crawler.settings.getint("CONCURRENT_REQUESTS")
-        self.context_launch_lock = asyncio.Lock()
-        self.context_kwargs: dict = settings.getdict("PLAYWRIGHT_CONTEXTS")
-        self.contexts: Dict[str, BrowserContext] = {}
-        self.context_semaphores: Dict[str, asyncio.Semaphore] = {
-            'default': asyncio.Semaphore(value=self.max_pages_per_context)
-        }
+        self.static_context_kwargs: dict = settings.getdict("PLAYWRIGHT_STATIC_CONTEXT")
+        self.dynamic_contexts_kwargs: [dict] = settings.getdict("PLAYWRIGHT_DYNAMIC_CONTEXTS") or []
+        self.context_semaphore: asyncio.Semaphore = asyncio.Semaphore(value=1)
 
         self.default_navigation_timeout: Optional[float] = None
         if "PLAYWRIGHT_DEFAULT_NAVIGATION_TIMEOUT" in settings:
@@ -128,10 +121,10 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         if settings.get("PLAYWRIGHT_ABORT_REQUEST"):
             self.abort_request = load_object(settings["PLAYWRIGHT_ABORT_REQUEST"])
 
-        self.user_agent_rotation_idx = 0
+        self.context_rotation_idx = 0
+        self.timeout_errors_in_row = 0
         self.page_load_until: Literal["commit", "domcontentloaded", "load", "networkidle"] = settings.get(
             "PLAYWRIGHT_PAGE_LOAD_UNTIL") or "load"
-        self.playwright_user_agents: list = crawler.settings.get("PLAYWRIGHT_USER_AGENTS") or []
         crawler.signals.connect(self._ip_changed, bhf_signals.ip_changed)
 
     @classmethod
@@ -140,70 +133,48 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
 
     def _engine_started(self) -> Deferred:
         """Launch the browser. Use the engine_started signal as it supports returning deferreds."""
-        if not self.playwright_user_agents:
-            return deferred_from_coro(self._launch_browser())
-        else:
-            return deferred_from_coro(self._launch_browser(user_agent=self.playwright_user_agents[1]))
+        return deferred_from_coro(self._launch_browser())
 
     def _ip_changed(self) -> Deferred:
         """Plan User-Agent rotation"""
-        return deferred_from_coro(self._plan_user_agent_rotation())
+        return deferred_from_coro(self._plan_context_rotation())
 
-    async def _launch_browser(self, user_agent=None) -> None:
+    async def _launch_browser(self) -> None:
         self.playwright_context_manager = PlaywrightContextManager()
         self.playwright = await self.playwright_context_manager.start()
         logger.info("Launching browser")
         browser_launcher = getattr(self.playwright, self.browser_type).launch
         self.browser = await browser_launcher(**self.launch_options)
         logger.info(f"Browser {self.browser_type} launched")
-        context_kwargs = self.context_kwargs
-        if user_agent is not None:
-            context_kwargs['user_agent'] = user_agent
-        self.contexts = {
-            'default': await self._create_browser_context('default', context_kwargs)
-        }
 
-    async def _create_browser_context(self, name: str, context_kwargs: dict) -> BrowserContext:
-        context = await self.browser.new_context(**context_kwargs)
-        context.on("close", self._make_close_browser_context_callback(name))
-        logger.debug("Browser context started: '%s'", name)
-        self.stats.inc_value("playwright/context_count")
-        if self.default_navigation_timeout is not None:
-            context.set_default_timeout(self.default_navigation_timeout)
-            context.set_default_navigation_timeout(self.default_navigation_timeout)
-        return context
-
-    async def _plan_user_agent_rotation(self) -> None:
+    async def _plan_context_rotation(self) -> None:
         logger.info("Re create browser: Before acquire")
-        await self.context_semaphores['default'].acquire()
+        await self.context_semaphore.acquire()
         logger.info("Re create browser: After acquire")
 
         await self._close()
-        if not self.playwright_user_agents:
-            await self._launch_browser()
-        else:
-            logger.info(f"Re create browser with new user-agent: "
-                        f"{self.playwright_user_agents[self.user_agent_rotation_idx]}")
-            await self._launch_browser(user_agent=self.playwright_user_agents[self.user_agent_rotation_idx])
-            self.user_agent_rotation_idx = self.user_agent_rotation_idx + 1
-            if self.user_agent_rotation_idx >= len(self.playwright_user_agents):
-                self.user_agent_rotation_idx = 0
+        await self._launch_browser()
+        self.context_rotation_idx = self.context_rotation_idx + 1
+        if self.context_rotation_idx >= len(self.dynamic_contexts_kwargs):
+            self.context_rotation_idx = 0
 
-        if self.context_semaphores['default'].locked():
-            self.context_semaphores['default'].release()
+        if self.context_semaphore.locked():
+            self.context_semaphore.release()
         logger.info("Re create browser: Completed")
 
     async def _create_page(self) -> Page:
         """Create a new page in a context, also creating a new context if necessary."""
-        context = self.contexts.get("default")
-
-        page = await context.new_page()
+        page_context_kwargs = self.static_context_kwargs
+        additional_page_context_kwargs = self.dynamic_contexts_kwargs[self.context_rotation_idx]
+        for key in additional_page_context_kwargs:
+            page_context_kwargs[key] = additional_page_context_kwargs[key]
+        page = await self.browser.new_page(page_context_kwargs)
         # await stealth_async(page)
         self.stats.inc_value("playwright/page_count")
         logger.debug(
             "[Context=%s] New page created, page count is %i (%i for all contexts)",
             "default",
-            len(context.pages),
+            1,
             self._get_total_page_count(),
         )
         if self.default_navigation_timeout is not None:
@@ -218,7 +189,7 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         return page
 
     def _get_total_page_count(self):
-        count = sum([len(context.pages) for context in self.contexts.values()])
+        count = 1
         current_max_count = self.stats.get_value("playwright/page_count/max_concurrent")
         if current_max_count is None or count > current_max_count:
             self.stats.set_value("playwright/page_count/max_concurrent", count)
@@ -228,10 +199,9 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
     def close(self) -> Deferred:
         yield super().close()
         yield deferred_from_coro(self._close())
-        self.context_semaphores.clear()
+        self.context_semaphore = None
 
     async def _close(self) -> None:
-        self.contexts.clear()
         if getattr(self, "browser", None):
             logger.info("Closing browser")
             await self.browser.close()
@@ -246,7 +216,7 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         page = None
         try:
             logger.info(f"[{request.url}] Before acquire")
-            await self.context_semaphores['default'].acquire()
+            await self.context_semaphore.acquire()
             logger.info(f"[{request.url}] After acquire")
             page = await self._create_page()
 
@@ -276,10 +246,15 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
             )
 
             result = await self._download_request_with_page(request, page)
+        except PlaywrightTimeoutError:
+            self.timeout_errors_in_row = self.timeout_errors_in_row + 1
+            logger.error(f"[{request.url}] Request timeout while downloading request")
+            return Response('', status=504)
         except:
             logger.exception(f"[{request.url}] Error while downloading request")
             return Response('', status=504)
         else:
+            self.timeout_errors_in_row = 0
             return result
         finally:
             try:
@@ -287,22 +262,25 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
                     try:
                         await asyncio.wait_for(page.close(), timeout=10)
                     except asyncio.TimeoutError:
-                        if self.context_semaphores['default'].locked():
-                            self.context_semaphores['default'].release()
-                        await self._plan_user_agent_rotation()
+                        if self.context_semaphore.locked():
+                            self.context_semaphore.release()
+                        await self._plan_context_rotation()
                     else:
-                        if self.context_semaphores['default'].locked():
-                            self.context_semaphores['default'].release()
+                        if self.context_semaphore.locked():
+                            self.context_semaphore.release()
+                        if self.timeout_errors_in_row > 3:
+                            self.timeout_errors_in_row = 0
+                            await self._plan_context_rotation()
                 else:
-                    if self.context_semaphores['default'].locked():
-                        self.context_semaphores['default'].release()
+                    if self.context_semaphore.locked():
+                        self.context_semaphore.release()
             except:
                 logger.exception(f"[{request.url}] Error on page close")
                 raise
 
     async def _download_request_with_page(self, request: Request, page: Page) -> Response:
         start_time = time()
-        response = await page.goto(request.url, wait_until=self.page_load_until)
+        response: PlaywrightResponse = await page.goto(request.url, wait_until=self.page_load_until)
 
         logger.info(f"[{request.url}] After goto")
         await self._apply_page_methods(page, request)
@@ -357,7 +335,6 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
                                 return
                             except Exception:
                                 timeouts += 1
-                                await page.screenshot(path=request.url, type="jpeg")
                                 if timeouts == group_length:
                                     raise
                                 else:
@@ -389,21 +366,6 @@ class ScrapyPlaywrightDownloadHandler(HTTPDownloadHandler):
         self.stats.inc_value(f"{stats_prefix}/resource_type/{response.request.resource_type}")
         self.stats.inc_value(f"{stats_prefix}/method/{response.request.method}")
 
-    def _make_close_page_callback(self, context_name: str) -> Callable:
-        def close_page_callback() -> None:
-            if context_name in self.context_semaphores and self.context_semaphores[context_name].locked():
-                self.context_semaphores[context_name].release()
-
-        return close_page_callback
-
-    def _make_close_browser_context_callback(self, name: str) -> Callable:
-        def close_browser_context_callback() -> None:
-            logger.debug("Browser context closed: '%s'", name)
-            if name in self.contexts:
-                self.contexts.pop(name)
-
-        return close_browser_context_callback
-
     def _make_request_handler(
             self, method: str, scrapy_headers: Headers, body: Optional[bytes], encoding: str = "utf8"
     ) -> Callable:
